
import os
import logging
import configparser

from RotL import windows as win

logger = logging.getLogger(__name__)

REMEDIATION_TYPES = ['files', 'process_names', 'scheduled_tasks', 'services', 'directories', 'pids', 'registry_keys', 'registry_values']

def delete_file(client, file_path):
    """Delete a file on a client.

    :param client: A lerc_api.Client object
    :param str file_path: Path to the file on the client.
    :return: A lerc_api.Command object    
    """
    cmd = client.Run(win.delete_file(file_path))
    logger.info("Issued command:{} to delete '{}' on LERC:{}".format(cmd.id, file_path, client.hostname))
    return cmd

def delete_registry_value(client, reg_path):
    """Delete a registry value on a client.

    :param client: A lerc_api.Client object
    :param str reg_path: A registry key path + '\\' + registry value
    :return: A lerc_api.Command object
    """
    cmd = client.Run(win.delete_registry_value(reg_path))
    logger.info("Issued command:{} to '{}' on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    return cmd

def delete_registry_key(client, reg_path):
    """Delete registry key on a client.

    :param lerc_api.Client client: A LERC Client.
    :param str reg_path: A registry key path that should be deleted (all values).
    :return: A lerc_api.Command
    """
    cmd = client.Run(win.delete_registry_key(reg_path))
    logger.info("Issued command:{} to '{}' on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    return cmd

def delete_service(client, service_name, service_path=None, auto_fill=False):
    """Delete a service from the registry.

    :param lerc_api.Client client: A LERC Client.
    :param str service_name: The name of the service that should be deleted.
    :param str service_path: (optional) Specify the path the the service for file deletion.
    :param bool auto_fill: If True, use wmic to query the client for PathName and State. Stop the service if running, delete the path, then delete the service from the registry.
    :return: A list of commands that were executed
    :rtype: list
    """
    commands = []
    if auto_fill:
        cmd = client.Run('wmic service where \'name like "{}"\' get name,pathname,state /format:list'.format(service_name))
        logger.info("Issued command:{} to get the PathName and State of service '{}' on LERC:{}".format(cmd.id, service_name, client.hostname))
        cmd.wait_for_completion()
        results = cmd.get_results(return_content=True)
        if results is None:
            logger.warn("Expected results for wmic service query command:{}".format(cmd.id))
        else:
            results = results.decode('utf-8')
            # for every none-empty line in the results string
            for line in [l for l in results.split('\n\n') if l]:
                if 'PathName' in line:
                    # get path from command line execution
                    if '"' in line:
                        service_path = line[len('PathName="'):]
                        service_path = service_path[:service_path.find('"')]
                    else:
                        service_path = line[len('PathName='):]
                        if service_path.find(' ') > 0:
                            # so there is a space without a '"'. lame. Assuming executable.
                            service_path = service_path[:service_path.find('.exe')+4]
                    logger.info("Found service PathName '{}'".format(service_path))
                elif 'State=Running' in line:
                    logger.info("Service is running on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    if service_path is not None:
        # either the user provided a service path or we got it with wmic via auto=True
        # go ahead and make the service file deletion a tuple of what evaluate_remediation_results needs
        commands.append((delete_file(client, service_path), 'files', service_path))
    cmd = client.Run(win.delete_service(service_name))
    logger.info("Issued command:{} to '{}' on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    commands.append(cmd)
    return commands

def delete_scheduled_task(client, task_name):
    """Delete a scheduled task.

    :param lerc_api.Client client: A LERC Client.
    :param str task_name: The name of the scheduled task.
    :return: A lerc_api.Command object
    """
    cmd = client.Run(win.delete_scheduled_task(task_name))
    logger.info("Issued command:{} to '{}' on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    return cmd

def delete_directory(client, dir_path):
    """Delete an entire directory.

    :param lerc_api.Client client: A LERC Client.
    :param str dir_path: The path to the directory.
    :return: A lerc_api.Command object
    """
    cmd = client.Run(win.delete_directory(dir_path))
    logger.info("Issued command:{} to delete '{}' on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    return cmd

def kill_process_name(client, process):
    """Kill all running processes by a process name.

    :param lerc_api.Client client: A LERC Client.
    :param str process: The process name
    :return: A lerc_api.Command object
    """
    cmd = client.Run(win.kill_process_name(process))
    logger.info("Issued command:{} to kill '{}' on LERC:{}".format(cmd.id, process, client.hostname))
    return cmd

def kill_process_id(client, pid):
    """Kill a process by its ID.

    :param lerc_api.Client client: A LERC Client.
    :param str pid: The process ID
    :return: A lerc_api.Command object
    """
    cmd = client.Run(win.kill_process_id(pid))
    logger.info("Issued command:{} to kill '{}' on LERC:{}".format(cmd.id, pid, client.hostname))
    return cmd

def evaluate_remediation_results(cmd, remediation_type, target_value):
    """Evaluate the remediation results for a given command by the type of remediation it is.

    :param lerc_api.Command cmd: The LERC Command to evaluate
    :param str remediation_type: The type of remediation the command represents.
    :param str target_value: The value of the artifact being remediated on the client.
    """
    if remediation_type not in REMEDIATION_TYPES:
        raise ValueError("Invalid remediation type '{}'. Valid types: {}".format(remediation_type, REMEDIATION_TYPES))

    logger.info("Waiting for command:{} - '{}' to COMPLETE..".format(cmd.id, cmd.command))
    cmd.wait_for_completion()
    results = cmd.get_results(return_content=True)
    if remediation_type == 'process_names':
        pname = target_value
        results = results.decode('utf-8') if results is not None else None
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error('Problem killing {} : {}:{}'.format(pname, cmd.status, error_message))
        elif 'SUCCESS' in results:
            logger.info("'{}' process names killed successfully : {}".format(pname, results))
        else:
            logger.warn("'{}' process names problem killing : {}".format(pname, results))
    elif remediation_type == 'pids':
        pid = target_value
        results = results.decode('utf-8') if results is not None else None
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error('Problem killing process id {}'.format(pid, cmd.status, error_message))
        elif 'SUCCESS' in results:
            logger.info("process id '{}' killed successfully : {}".format(pid, results))
        else:
            logger.warn("problem killing process id '{}' : {}".format(pid, results))
    elif remediation_type == 'registry_values':
        rkey = target_value
        results = results.decode('utf-8') if results is not None else None
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error("Problem deleting '{}'".format(rkey, cmd.status, error_message))
        elif 'success' not in results:
            logger.warn("Problem deleting '{}' : {}".format(rkey, results))
        else:
            logger.info("Deleted '{}' : {}".format(rkey, results))
    elif remediation_type == 'registry_keys':
        rkey = target_value
        results = results.decode('utf-8') if results is not None else None
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error("Problem deleting '{}'".format(rkey, cmd.status, error_message))
        elif 'success' not in results:
            logger.warn("Problem deleting '{}' : {}".format(rkey, results))
        else:
            logger.info("Deleted '{}' : {}".format(rkey, results))
    elif remediation_type == 'files':
        fname = target_value
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error("Problem deleting '{}'".format(fname, cmd.status, error_message))
        if results is not None:
            logger.warn("Problem deleting '{}' : {}".format(fname, results.decode('utf-8')))
        else:
            logger.info("File '{}' deleted successfully.".format(fname))
    elif remediation_type == 'directories': 
        directory = target_value
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error("Problem deleting directory '{}' : {}".format(directory, cmd.status, error_message))
        elif results is not None:
            logger.warn("Problem deleting directory '{}' : {}".format(directory, results.decode('utf-8')))
        else:
            logger.info("Successfully deleted '{}'.".format(directory))
    elif remediation_type == 'scheduled_tasks':
        task = target_value
        results = results.decode('utf-8') if results is not None else None
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error("Problem deleting scheduled task '{}' : {}: {}".format(task, cmd.status, error_message))
        elif 'SUCCESS' in results:
            logger.info("Scheduled task '{}' deleted successfully : {}".format(task, results))
        else:
            logger.warn("Problem deleting scheduled task '{}' : {}".format(task, results))
    elif remediation_type == 'services':
        service_name = target_value
        results = results.decode('utf-8') if results is not None else None
        if cmd.status != 'COMPLETE':
            error_message = cmd.get_error_report()['error']
            logger.error("Problem remediating service '{}' : {}: {}".format(service_name, cmd.status, error_message))
        elif 'service was stopped successfully' in results:
            logger.info("The '{}' service was stopped successfully.".format(service_name))
        elif 'DeleteService SUCCESS' in results:
            logger.info("The '{}' service was successfully deleted from the registry.".format(service_name))
        else:
            logger.warn("Problem remediating service '{}' : command:{}".format(service_name, cmd.id))
    else:
        logger.error("Unexpected remediation evaluation : command_id={} - {} - {}".format(cmd.id, remediation_type, target_value))


def Remediate(client, remediation_script):
    """Perform a remediation effort on a client via a remediation configuration script.

    :param lerc_api.Client client: The LERC Client
    :param str remediation_script: The path to the remediation config file.
    """
    if not os.path.exists(remediation_script):
        logger.error("'{}' Does not exist".format(remediation_script))
        return False
    logger.info("Attempting to perform mass remediation via '{}' on LERC:{}".format(remediation_script, client.hostname))

    config = configparser.ConfigParser()
    config.read(remediation_script)

    commands = {'files': [],
                'process_names': [],
                'scheduled_tasks': [],
                'directories': [],
                'pids': [],
                'registry_keys': [],
                'registry_values': [],
                'services': []}

    # Order matters
    processes = config['process_names']
    for p in processes:
        commands['process_names'].append(kill_process_name(client, processes[p]))

    pids = config['pids']
    for p in pids:
        commands['pids'].append(kill_process_id(client, pids[p]))

    regValues = config['registry_values']
    for key in regValues:
        commands['registry_values'].append(delete_registry_value(client, regValues[key]))

    regKeys = config['registry_keys']
    for key in regKeys:
        commands['registry_keys'].append(delete_registry_key(client, regKeys[key]))

    files = config['files'] 
    for f in files:
        commands['files'].append(delete_file(client, files[f]))

    dirs = config['directories']
    for d in dirs:
        commands['directories'].append(delete_directory(client, dirs[d]))

    tasks = config['scheduled_tasks']
    for t in tasks:
        commands['scheduled_tasks'].append(delete_scheduled_task(client, tasks[t]))

    services = config['services']
    for s in services:
        commands['services'].append(delete_service(client, services[s]))

    # evaluate results and report
    for cmd in commands['process_names']:
        cmd_pname = cmd.command[cmd.command.find('"')+1:cmd.command.rfind('"')]
        # get the process name that is killed in this command, should be single results
        pname = [p for p in [processes[p] for p in processes] if p == cmd_pname][0]
        evaluate_remediation_results(cmd, 'process_names', pname)

    for cmd in commands['pids']:
        pid = [p for p in [pids[p] for p in pids] if p in cmd.command][0]
        evaluate_remediation_results(cmd, 'pids', pid)

    for cmd in commands['registry_values']:
        # get rkey that has path and key in cmd.command
        _cmd_str = cmd.command
        rvalue = [r for r in [regValues[r] for r in regValues] if r[r.rfind('\\')+1:] in _cmd_str and r[:r.rfind('\\')] in _cmd_str][0]
        evaluate_remediation_results(cmd, 'registry_values', rvalue)

    for cmd in commands['registry_keys']:
        rkey = [r for r in [regKeys[r] for r in regKeys] if r in cmd.command][0]
        evaluate_remediation_results(cmd, 'registry_keys', rkey)

    for cmd in commands['files']:
        fname = [f for f in [files[f] for f in files] if f in cmd.command][0]
        evaluate_remediation_results(cmd, 'files', fname)

    for cmd in commands['directories']:
        directory = [d for d in [dirs[d] for d in dirs] if d in cmd.command][0]
        evaluate_remediation_results(cmd, 'directories', directory)

    for cmd in commands['scheduled_tasks']:
        task = [t for t in [tasks[t] for t in tasks] if t in cmd.command][0]
        evaluate_remediation_results(cmd, 'scheduled_tasks', task)

    # services are a tuple list of commands
    for cmds in commands['services']:
        for cmd in cmds:
            if isinstance(cmd, tuple):
                evaluate_remediation_results(cmd[0], cmd[1], cmd[2])
            else:
                service_name = [s for s in [services[s] for s in services] if s in cmd.command][0]
                evaluate_remediation_results(cmd, 'services', service_name)



import os
import time
import types
import logging
import progressbar

from lerc_control.lerc_api import CONFIG, check_config

# Get the working lerc_control directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

logger = logging.getLogger("lerc_control."+__name__)

class TablePrinter(object):
    """Print a list of dicts as a table.

    :param fmt: list of tuple(heading, key, width)
    :param sep: string, separation between columns
    :param ul: string, character to underline column label, or None for no underlining
    :return: A string representation of the table, ready to be printed

    Each tuple in the fmt list of tuples is like so:

    :heading: str, column label
    :key: dictionary key to value to print
    :width: int, column width in chars
    """
    def __init__(self, fmt, sep=' ', ul=None):
        super(TablePrinter,self).__init__()
        self.fmt   = str(sep).join('{lb}{0}:{1}{rb}'.format(key, width, lb='{', rb='}') for heading,key,width in fmt)
        self.head  = {key:heading for heading,key,width in fmt}
        self.ul    = {key:str(ul)*width for heading,key,width in fmt} if ul else None
        self.width = {key:width for heading,key,width in fmt}

    def row(self, data):
        return self.fmt.format(**{ k:str(data.get(k,''))[:w] for k,w in self.width.items() })

    def __call__(self, dataList):
        _r = self.row
        res = [_r(data) for data in dataList]
        res.insert(0, _r(self.head))
        if self.ul:
            res.insert(1, _r(self.ul))
        return '\n'.join(res)


def wait_for_completion_and_display_any_transfer_progress(cmd, desc=None):
    if cmd.operation.lower() not in ['download', 'upload']:
        logger.info("Not a transfer command.")
        return cmd.wait_for_completion()
    if desc is None:
        desc = "CMD:{} - {} progress: [".format(cmd.id, cmd.operation)
    if not cmd.started:
        logger.info("Waiting for command to move from {} to STARTED..".format(cmd.status))
    while not cmd.started:
        if not cmd.progress():
            return cmd.wait_for_completion()
        time.sleep(1)

    if cmd.started and cmd.filesize == 0:
        logger.warning("Unexpected command state and filesize combination.")
        return cmd.wait_for_completion()
    logger.info("Command {} has STARTED ..".format(cmd.id))

    pbar = progressbar.ProgressBar(maxval=cmd.filesize, widgets=[progressbar.Bar("=", desc, ']'), ' ', progressbar.Percentage()])
    pbar.term_width = int(pbar.term_width / 2)

    pbar.start()
    while not cmd.complete:
        pbar.update(cmd.file_position)
        if not cmd.progress():
            return cmd.wait_for_completion()
        time.sleep(0.5)
    pbar.finish()
    return True

def install_rekall(lerc, rekall_installer_path=None, rekall_install_command=None):
    """Install rekall on a client.
    """
    required_info = []
    if rekall_installer_path is None:
        required_info.append('rekall_installer_path')
    if rekall_install_command is None:
        required_info.append('rekall_install_command')

    if rekall_installer_path is not None:
        if not os.path.exists(rekall_installer_path):
            if rekall_installer_path[0] == '/':
                rekall_installer_path = BASE_DIR + rekall_installer_path
            else:
                rekall_installer_path = BASE_DIR + '/' + rekall_installer_path
        if not os.path.exists(rekall_installer_path):
            logger.error("Path to rekall installer does not exist: {}".format(rekall_installer_path))
            return False

    config = check_config(CONFIG, required_keys=required_info)
    if not config:
        logger.error("Missing rekall config items.")
        return False
        
    if rekall_install_command is None:
        rekall_install_command = config['default']['rekall_install_command'] if config.has_option('default', 'rekall_install_command') else rekall_install_command
        if lerc.profile != "default":
            rekall_install_command = config[lerc.profile]['rekall_install_command'] if config.has_option(lerc.profile, 'rekall_install_command') else rekall_install_command
    if rekall_installer_path is None:
        rekall_installer_path = config['default']['rekall_installer_path'] if config.has_option('default', 'rekall_installer_path') else rekall_installer_path
        if lerc.profile != "default":
            rekall_installer_path = config[lerc.profile]['rekall_installer_path'] if config.has_option(lerc.profile, 'rekall_installer_path') else rekall_installer_path

    downcmd = lerc.Download(rekall_installer_path, client_file_path='rekall_installer.exe')
    logger.info("Issued {} to download rekall_installer.exe on {}".format(downcmd.id, lerc.hostname))
    installcmd = lerc.Run(rekall_install_command)
    logger.info("Issued {} to install rekall on {}".format(installcmd.id, lerc.hostname))
    installcmd.wait_for_completion()
    return True


#!/usr/bin/python3

import os
import sys
import time
import argparse
import logging
import coloredlogs
import pprint

import lerc_api

from configparser import ConfigParser

# configure logging #
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - [%(levelname)s] %(message)s')
# set noise level
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
logging.getLogger('lerc_api').setLevel(logging.INFO)

logger = logging.getLogger(__name__)
coloredlogs.install(level='INFO', logger=logger)


def upgrade_host(hostname, upgrade_bat_path, lerc_msi_path):
    """Have a lerc upgrade itself to a newer version.

    :param str hostname: The hostname to issue upgrade commands.
    :param str upgrade_bat_path: Path to the upgrade.bat file used to manage the bootstrapped upgrade.
    :param str lerc_msi_path: Path to the current version of the lerc we're upgrading the client to.

    Upgrade Steps:

      #. Drop lercSetup.msi
      #. Drop upgrade.bat 
      #. Execute upgrade.bat async=True with correct params ex. -> upgrade.bat 0 15 2048 "https://your-server-address/"  -> company=0 reconnectdelay=15 chunksize=2048 serverurls="https://your-server-address/"
      #. Issue Quit command to host
    """

    host_commands = []

    ls = lerc_api.lerc_session()
    host = ls.get_host(hostname)
    if not host:
        logging.error("'{}' does not exist.".format(hostname))
        return "ERROR: '{}' does not exist.".format(hostname)
    logger = logging.getLogger(__file__+".upgrade_host")
    logger.info("Issuing upgrade commands to {}".format(hostname))

    # delete any existing lercSetup.msi that might already be on the host
    result = host.Run("DEL lercSetup.msi")
    host_commands.append(result)

    file_name = lerc_msi_path[lerc_msi_path.rfind('/')+1:]
    result = host.Download(file_name, client_file_path=file_name, analyst_file_path=lerc_msi_path)
    host_commands.append(result)

    file_name = upgrade_bat_path[upgrade_bat_path.rfind('/')+1:]
    result = host.Download(file_name, client_file_path=file_name, analyst_file_path=upgrade_bat_path)
    host_commands.append(result)

    run_cmd = config['default']['upgrade_cmd']
    result = host.Run(run_cmd.format(host.id), async=True)
    host_commands.append(result)

    result = host.Quit()
    host_commands.append(result)
    return host_commands


def compare_version_strings(client, production_lerc_version):
    """Check the version string the server has on a client with the current production version string. Note: version format  is #.#.#.#

    :param dict client: A client dictionary obtained from the lerc server.
    :param str production_lerc_version: A string representation of the current production lerc version.
    :return: True if the client version is older than the production version; False if equal or greater.
    """
    logger = logging.getLogger(__name__+".compare_versions")
    logger.debug("Version comparision : Client Current Version={} - Production Version={}".format(client.version, production_lerc_version))
    production_lerc_version_str = production_lerc_version
    production_lerc_version = production_lerc_version.split('.')
    if client.version == production_lerc_version_str:
        logger.warn("Client '{}' is already at the current production version '{}'".format(client.hostname, production_lerc_version_str))
        return False
    else:
        client_ver = client.version.split('.')
        if len(client_ver) != len(production_lerc_version):
            logger.warn("Version annomoly on {} - client version:{} and production version:{}".format(client.hostname, client.version, production_lerc_version_str))
            return False
        for i in range(len(client_ver)):
            if client_ver[i] < production_lerc_version[i]:
                return True
        else:
            logger.error("Client version:{} greater than production_lerc_version:{}. Is your configuration wrong?".format(client.version, production_lerc_version_str))
            return False


if __name__ == "__main__":

    config = lerc_api.load_config(required_keys=['upgrade_cmd', 'upgrade_bat', 'client_installer', 'production_lerc_version'])
    env_choices = [c for c in config.sections() if 'default' not in c and '_' not in c]

    parser = argparse.ArgumentParser(description="A script to upgrade clients with a new lerc version.")
    parser.add_argument('-f', '--file', action="store", help="specify the path to a lercSetup.msi (default used from config file)")
    parser.add_argument('-d', '--debug', action="store_true", help="set logging to DEBUG")
    parser.add_argument('-e', '--environment', choices=env_choices, action='store', help="only upgrade hosts in a certain environment.")
    parser.add_argument('-c', '--client-hostname', action='store', help="upgrade this specific client, only")
    parser.add_argument('--force', action='store_true', help="make the client upgrade regardless of any other conditions, such as production_lerc_version")
    parser.add_argument('--write-commands', action='store_true', help="write all of the issued commands to a file for future reference.")

    args = parser.parse_args()

    if args.debug:
        logging.getLogger('lerc_api').setLevel(logging.DEBUG)
        coloredlogs.install(level='DEBUG', logger=logger)
        logger.setLevel(logging.DEBUG)

    upgrade_bat_path = lerc_msi_path = production_lerc_version = company_id = None
    # It's possible different environments have different requirements
    if args.environment:
        if config.has_option(args.environment, 'upgrade_bat'):
            upgrade_bat_path = config[args.environment]['upgrade_bat']
        if config.has_option(args.environment, 'client_installer'):
            lerc_msi_path = config[args.environment]['client_installer']
        if config.has_option(args.environment, 'production_lerc_version'):
            production_lerc_version = config[args.environment]['production_lerc_version']
        if config.has_option(args.environment, 'company_id'):
            company_id = config[args.environment].getint('company_id')

    # defaults
    if upgrade_bat_path is None:
        upgrade_bat_path = config['default']['upgrade_bat']
    if lerc_msi_path is None:
        lerc_msi_path = config['default']['client_installer']
    if production_lerc_version is None:
        production_lerc_version = config['default']['production_lerc_version']

    logger.debug("upgrade_bat_path: {}".format(upgrade_bat_path))
    logger.debug("lerc_msi_path: {}".format(lerc_msi_path))
    logger.debug("production_lerc_version: {}".format(production_lerc_version))

    # check paths
    for path in [upgrade_bat_path, lerc_msi_path]:
        if not os.path.exists(path):
            logger.error("Does not exist: {}".format(path))
            sys.exit(1)

    ls = lerc_api.lerc_session()

    if args.client_hostname:
        client = ls.get_host(args.client_hostname)
        if not client:
            logger.error("Client doesn't exist")
            sys.exit(1)
        # make sure this host's company_id matches what we have for this environment
        if args.environment:
            if company_id is None:
                logger.error("Company id is not specified in the configuration for '{}'".format(args.environment))
                sys.exit(1)
            elif company_id != int(client.company_id):
                logger.warn("Client company id and company id for '{}' do not match. Wrong hostname?".format(args.environment))
                sys.exit(1)
        if not args.force:
            # check lerc version strings
            if client.version is not None:
                proceed = compare_version_strings(client, production_lerc_version)
                if not proceed: # we exit since we're only working on one host
                    sys.exit(1)
        commands = upgrade_host(args.client_hostname, upgrade_bat_path, lerc_msi_path)
        if args.write_commands:
            with open(args.client_hostname+"_upgrade.log", 'w') as fh:
                fh.write(pprint.pformat(commands))
            print("Wrote client upgrade commands to '{}'".format(args.client_hostname+"_upgrade.log"))
        sys.exit()

    host_commands = {}
    for host in ls.yield_hosts():
        if host.hostname == 'WIN-1TMIV79KTI8' or host.hostname == 'icinga' \
                                                 or host.hostname == 'W7GOTCHAPC':
            continue
        if args.environment:
            if company_id is None:
                logger.error("Company id is not specified in the configuration for '{}'".format(args.environment))
                sys.exit(1)
            elif company_id != int(host.company_id):
                logger.debug("Client company id and company id for '{}' do not match for {}/{}".format(args.environment, host.hostname, host.id))
                continue
        if host.status != 'UNINSTALLED':
            if not compare_version_strings(host, production_lerc_version):
                continue
            host_commands[host.hostname] = upgrade_host(host.hostname, upgrade_bat_path, lerc_msi_path)

    if args.write_commands:
        with open("lerc_upgrade_commands.txt", 'w') as fh:
            fh.write(pprint.pformat(host_commands))
        print("Wrote lerc_upgrade_commands.txt.")

    sys.exit()



#!/usr/bin/env python3

import os
import sys
import time
import argparse
import logging
import coloredlogs
import pprint
from lerc_control import lerc_api, collect
from lerc_control.scripted import execute_script, get_script_results, load_script, COMMON_CLEANUP_COMMANDS
from lerc_control.helpers import TablePrinter

BASE_DIR = os.path.dirname(os.path.realpath(__file__))

# configure logging #
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - [%(levelname)s] %(message)s')
# set noise level
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
logging.getLogger('lerc_api').setLevel(logging.INFO)
logging.getLogger('lerc_control').setLevel(logging.INFO)
#logging.getLogger('lerc_control.scripted').setLevel(logging.INFO)
#logging.getLogger('lerc_control.collect').setLevel(logging.INFO)

logger = logging.getLogger('lerc_ui')
coloredlogs.install(level='INFO', logger=logger)


def _parse_collection_scripts(collect_scripts):
    """
    Parse list of collection scripts names and resturn a list structure formatted to create
    argparse arguments. Where, list elements are arguments presented as dictionaries like so:
        ```
        {script: raw collection script name,
        abrv: argparse abbreviation,
        name: argument name}
        ```
    If 'subparser' is a dict key, the value will be the name of the subparser and a 'args' key will
    contain a list of arguments for that subparser.
    """
    arguments = []
    subs = [name[len('collect_'):name.find('_', len('collect_'))] for name in collect_scripts if name.count('_') > 1]
    subs = set([name for name in subs if subs.count(name)>1])
    for sub in subs:
        arguments.append({ 'subparser': sub,
                            'args': []})
    for script in collect_scripts:
        
        name = script[len('collect_'):]
        sub_name = None
        if name.count('_') > 0:
            sub_name = name[:name.find('_')]
        abrv = name[0]
        # can't use '-h' help arg
        if abrv == 'h':
            abrv = 'c' + abrv
        tmp = name
        for i in range(name.count('_')):
            tmp = tmp[tmp.find('_')+1:]
            abrv += tmp[0]
        argument = {'script': script, 
                    'name': name,
                    'abrv': abrv}
        if sub_name in subs:
            for arg in arguments:
                if 'subparser' in arg and arg['subparser'] == sub_name:
                    argument['name'] = name[len(sub_name)+1:]
                    argument['abrv'] = abrv[1:]
                    arg['args'].append(argument)
        else:
            arguments.append(argument)
    return arguments


def main():

    parser = argparse.ArgumentParser(description="User interface to the LERC control server")
    # LERC environment choices
    config = lerc_api.load_config()
    env_choices = [ sec for sec in config.sections() if config.has_option(sec, 'server') ]

    parser.add_argument('-e', '--environment', action="store", default='default', 
                        help="specify an environment to work with. Default='default'", choices=env_choices)
    parser.add_argument('-d', '--debug', action="store_true", help="set logging to DEBUG", default=False)
    parser.add_argument('-c', '--check', action="store", help="check on a specific command id")
    parser.add_argument('-r', '--resume', action='store', help="resume a pending command id") 
    parser.add_argument('-g', '--get', action='store', help="get results for a command id")
    parser.add_argument('-k', '--cancel', action='store', help="Tell the server to CANCEL this command.")

    subparsers = parser.add_subparsers(dest='instruction') #title='subcommands', help='additional help')

    # Query
    parser_query = subparsers.add_parser('query', help="Query the LERC Server")
    parser_query.add_argument('query', help="The search you want to run. Enter 'fields' to see query fields.")
    parser_query.add_argument('-rc', '--return-commands', action='store_true', help="Return command results (even if no cmd fields specified)")
 
    # Initiate new LERC commands
    parser_run = subparsers.add_parser('run', help="Run a shell command on the host.")
    parser_run.add_argument('hostname', help="the host you'd like to work with")
    parser_run.add_argument('command', help='The shell command for the host to execute`')
    parser_run.add_argument('-a', '--async', action='store_true', help='Set asynchronous to true (do NOT wait for output or command to complete)')
    parser_run.add_argument('-p', '--print-only', action='store_true', help='Only print results to screen.')
    parser_run.add_argument('-w', '--write-only', action='store_true', help='Only write results to file.')
    parser_run.add_argument('-o', '--output-filename', default=None, action='store', help='Specify the name of the file to write any results to.')

    parser_upload = subparsers.add_parser('upload', help="Upload a file from the client to the server")
    parser_upload.add_argument('hostname', help="the host you'd like to work with")
    parser_upload.add_argument('file_path', help='the file path on the client')

    parser_download = subparsers.add_parser('download', help="Download a file from the server to the client")
    parser_download.add_argument('hostname', help="the host you'd like to work with")
    parser_download.add_argument('file_path', help='the path to the file on the server')
    parser_download.add_argument('-f', '--local-file', help='where the client should write the file')

    parser_quit = subparsers.add_parser('quit', help="tell the client to uninstall itself")
    parser_quit.add_argument('hostname', help="the host you'd like to work with")

    parser_script = subparsers.add_parser('script', help="run a scripted routine on this lerc.")
    parser_script.add_argument('hostname', help="the host you'd like to work with")
    parser_script.add_argument('-l', '--list-scripts', action='store_true', help="list scripts availble to lerc_ui")
    parser_script.add_argument('-s', '--script-name', help="provide the name of a script to run")
    parser_script.add_argument('-f', '--file-path', help="the path to a custom script you want to execute")

    ## response functions - collect, contain, and remediate

    # collect
    parser_collect = subparsers.add_parser('collect', help="Default (no arguments): perform a full lr.exe collection")
    parser_collect.add_argument('-d', '--directory', action='store', help="Compress contents of a client directory and collect")
    # NOTE for file-path gets picked up by upload parser and handed of as basic upload command
    parser_collect.add_argument('-f', '--file-path', action='store', help="Path to file on client you want to collect")
    parser_collect.add_argument('-rkv', '--reg-key-value', action='store', help="Collet the registry data at the specific registry key value path. Format:\
                                                                               HKLM\\software\\microsoft\\windows\\currentversion\\run\\SPECIFIC-KEY")
    parser_collect.add_argument('-rk', '--reg-key', action='store', help="Collet all registry value data at the specific registry key path. Format:\
                                                                               HKLM\\software\\wow6432node\\microsoft\\windows\\currentversion\\run")
    parser_collect.add_argument('-mc', '--multi-collect', action='store', help="Path to a multiple collection file")
    parser_collect.add_argument('-pm', '--process-memory', action='store', help='Use procdump to collect memory for this PID')
    parser_collect.add_argument('-pms','--proc-mem-strings', action='store', help='Use strings2 to collect process memory strings for this PID')
    parser_collect.add_argument('hostname', help="the host you'd like to work with")
    # Make collect argument options of collection scripts
    collect_scripts = []
    if config.has_section('scripts'):
        collect_scripts = [sname for sname in config['scripts'] if sname.startswith('collect_')]
    arguments = _parse_collection_scripts(collect_scripts)
    collect_subparsers = None
    sub_collect_parsers = {}
    for arg in arguments:
        if 'subparser' in arg:
            if collect_subparsers is None:
                collect_subparsers = parser_collect.add_subparsers(dest='sub_collect_scipts')
            subname = arg['subparser']
            sub_collect_parsers[subname] = collect_subparsers.add_parser(subname, help="Default: perform all {} collections.".format(subname),
                                                                         description='All options will be set if no option is specified.')
            for subarg in arg['args']:
                name = subarg['name']
                script = subarg['script']
                abrv =  subarg['abrv']
                script_config = load_script(config['scripts'][script])
                description = script.replace('_',' ')
                if script_config and script_config.has_section('overview'):
                    description = script_config['overview']['description'] if script_config.has_option('overview', 'description') else description
                # try to make sure we don't add the same option string more than once
                if '-'+abrv in sub_collect_parsers[subname]._option_string_actions:
                    #print(sub_collect_parsers[subname]._option_string_actions['-'+abrv])
                    abrv = abrv+name[1]
                sub_collect_parsers[subname].add_argument('-{}'.format(abrv), '--{}'.format(name), dest='{}'.format(script), action='store_true', help=description)
        else:
            abrv = arg['abrv']
            name = arg['name']
            script = arg['script']
            script_config = load_script(config['scripts'][script])
            description = script.replace('_',' ')
            if script_config and script_config.has_section('overview'):
                description = script_config['overview']['description'] if script_config.has_option('overview', 'description') else description
            parser_collect.add_argument('-{}'.format(abrv), '--{}'.format(name), dest='{}'.format(script), action='store_true', help=description)

    parser_contain = subparsers.add_parser('contain', help="Contain an infected host")
    parser_contain.add_argument('hostname', help="the host you'd like to work with")
    parser_contain.add_argument('-on', action='store_true', help="turn on containment")
    parser_contain.add_argument('-off', action='store_true', help="turn off containment")
    parser_contain.add_argument('-s', '--status', action='store_true', help="Get containment status of host")

    parser_remediate = subparsers.add_parser('remediate', help="Remediate an infected host")
    parser_remediate.add_argument('hostname', help="the host you'd like to work with")
    parser_remediate.add_argument('--write-template', action='store_true', default=False, help='write the remediation template file as remediate.ini')
    parser_remediate.add_argument('-f', '--remediation-file', help='the remediation file describing the infection')
    parser_remediate.add_argument('-drv', '--delete-registry-value', help='delete a registry value and all its data')
    parser_remediate.add_argument('-drk', '--delete-registry-key', help='delete all values at a registry key path')
    parser_remediate.add_argument('-df', '--delete-file', help='delete a file')
    parser_remediate.add_argument('-kpn', '--kill-process-name', help='kill all processes by this name')
    parser_remediate.add_argument('-kpid', '--kill-process-id', help='kill process id')
    parser_remediate.add_argument('-dd', '--delete-directory', help='Delete entire directory')
    parser_remediate.add_argument('-ds', '--delete-service', help='Delete a service from the registry and the ServicePath from the file system.')
    parser_remediate.add_argument('-dst', '--delete-scheduled-task', help='Delete a scheduled task by name')

    args = parser.parse_args()

    if args.debug:
        logging.getLogger('lerc_api').setLevel(logging.DEBUG)
        logging.getLogger('lerc_control').setLevel(logging.DEBUG)
        coloredlogs.install(level='DEBUG', logger=logger)

    # our local lerc_session
    ls = lerc_api.lerc_session(profile=args.environment)

    # root options
    if args.check:
        command = ls.get_command(args.check)
        if not command:
            sys.exit()
        print(command)
        if command.status == 'ERROR':
            print("ERROR Report:")
            pprint.pprint(command.get_error_report(), indent=5)
        sys.exit()
    elif args.get:
        logger.info("Asking for command results for CMD={}".format(args.get))
        command = ls.get_command(args.get)
        if command:
            logger.info("Getting results for {} command id {} issued to {}".format(command.operation, command.id, command.hostname))
            command.get_results(chunk_size=16384)
            print(command)
        sys.exit()
    elif args.resume:
        command = ls.get_command(args.resume)
        command.wait_for_completion()
        if command:
            print(command)
        sys.exit()
    elif args.cancel:
        cmds = args.cancel.split(',')
        # enumerate all commands that span CMD_ID-CMD_ID
        for cmd in cmds:
            if isinstance(cmd, str) and '-' in cmd:
                if cmd.count('-') != 1:
                    logger.error("Incorrect command id format string. Must be: 'int' OR 'int,int' OR 'int-int' OR 'int[,int,int-int]'")
                    sys.exit(1)
                first, second = cmd.split('-')[:2]
                # +1 to make it inclusive
                cmds.extend(list(range(int(first), int(second)+1)))
                cmds.remove(cmd)
        for cmd in cmds:
            command = ls.get_command(cmd)
            if command:
                command.cancel()
                print(command)
        sys.exit()

    if args.instruction == 'query':
        if args.query == 'fields':
            print("\nAvailable query fields:\n")
            fmt = [ ('Field', 'field', 14),
                    ('Description', 'description', 80) ]
            print( TablePrinter(fmt, sep='  ', ul='=')(lerc_api.QUERY_FIELD_DESCRIPTIONS) )
            print()
            print("NOTE:")
            print("  1) Fields are ANDed by default. Fields can be negated by appending '-' or '!' to the front of the field (no space) or by specifying 'NOT ' in front of the field (space).")
            print("  2) A leading '-' with no space in front will cause the argument parser to misinterpret the query as a command line argument option.")
            print() 
            sys.exit()
        query = lerc_api.parse_lerc_server_query(args.query)
        if args.return_commands: 
            query['rc'] = True
        results = ls.query(**query)
        if not results:
            sys.exit()
        clients = [ c.get_dict for c in results['clients']]
        fmt = [ ('ID', 'id', 5),
                ('Hostname', 'hostname', 20),
                ('Status', 'status', 11),
                ('Version', 'version', 8),
                ('Sleep Cycle', 'sleep_cycle', 11),
                ('Install Date', 'install_date', 24),
                ('Last Activity', 'last_activity', 24),
                ('Company ID', 'company_id', 10)]
        print("\nClient Results:\n")
        print( TablePrinter(fmt, sep='  ', ul='=')(clients))
        print("Total Client Results:{}".format(len(clients)))
        print()
        commands = [ c.get_dict for c in results['commands']]
        if commands:
            fmt = [ ('ID', 'command_id', 9),
                    ('Client ID', 'client_id', 9),
                    ('Hostname', 'hostname', 20),
                    ('Operation', 'operation', 11),
                    ('Status', 'status', 9),
                    ('Evaluated Time', 'evaluated_time', 24)]
                  #  ('Version', 'version', 8),
                  #  ('Sleep Cycle', 'sleep_cycle', 11),
                  #  ('Install Date', 'install_date', 20),
                  #  ('Last Activity', 'last_activity', 20)]
            print("\nCommand Results:\n")
            print( TablePrinter(fmt, sep='  ', ul='=')(commands))
            #for cmd in commands:
            #    print(cmd)
            print() 
        sys.exit()

    # if we're here, then an instructions been specified and the args.hostname is a thing
    client = ls.get_host(args.hostname)
    if isinstance(client, list):
        logger.critical("More than one result. Not handled yet..")
        for c in client:
            print(c)
            print()
        sys.exit(1) 

    # Auto-Deployment Jazz
    bad_status = False
    if client:
        if client.status == 'UNINSTALLED' or client.status == 'UNKNOWN':
            logger.info("Non-working client status : {}".format(client.status))
            bad_status = True

    if not client or bad_status:
        config = ls.get_config
        if config.has_option('default', 'cb_auto_deploy') and not config['default'].getboolean('cb_auto_deploy'):
            logger.info("CarbonBlack auto-deployment turned off. Exiting..")
            sys.exit(0)
        logger.info("Attempting to deploy lerc with CarbonBlack..")
        try:
            from cbapi import auth
            from lerc_control.deploy_lerc import deploy_lerc, CbSensor_search
        except:
            logger.error("Failed to import deployment functions from lerc_control.deploy_lerc OR cbapi.")
            sys.exit(1)
        logging.getLogger('lerc_control.deploy_lerc').setLevel(logging.ERROR)
        environments = auth.FileCredentialStore("response").get_profiles()
        sensors = []
        logger.debug("Trying to find the sensor in the available carbonblack environments.")
        for env in environments:         
            sensor = CbSensor_search(env, args.hostname)
            if sensor:
                logger.debug("Found sensor in {} environment".format(env))
                sensors.append((env, sensor))
        if len(sensors) > 1:
            logger.warning("A CarbonBlack Sensor was found by that hostname in multiple environments.")
            analyst_options = {}
            print("\nMultiple Sensors found:\n")
            i = 0
            for s in sensors:
                i+=1
                analyst_options[i] = s[0]
                print("------------\n\tEnvironment: {}".format(s[0]))
                print("\tID: {}".format(s[1].id))
                print("\tStatus: {}".format(s[1].status))
                print("\tHostname: {}".format(s[1].computer_name))
                print("\tDNS Name: {}".format(s[1].computer_dns_name))
                print("\tOS Type: {}".format(s[1].os_environment_display_string))
                print()
            print("\nWhich environment do you want to use?")
            for num, env in analyst_options.items():
                print("\t{}) {}".format(num, env))
            choice = None
            while choice is None:
                i = int(input("Enter the number corresponding to the environment: "))
                if i not in analyst_options.keys():
                    print("Incorrect choice.")
                else:
                    choice = i
            sensors = [s for s in sensors if s[0] == analyst_options[choice]]
            print("proceeding with {}".format(sensors[0]))

        if sensors:
            logging.getLogger('lerc_control.deploy_lerc').setLevel(logging.INFO)
            sensor = sensors[0][1]
            config = lerc_api.check_config(config, required_keys=['lerc_install_cmd', 'client_installer'])
            result = deploy_lerc(sensor, config[sensors[0][0]]['lerc_install_cmd'], lerc_installer_path=config['default']['client_installer'])
            if result: # modify deploy_lerc to use new client objects
                logger.info("Successfully deployed lerc to this host: {}".format(args.hostname))
                client = ls.get_host(args.hostname)
        else:
            logger.error("Didn't find a sensor in CarbonBlack by this hostname")
            sys.exit(0)
 
    # remediation
    if args.instruction == 'remediate':
        if not args.debug:
            logging.getLogger('lerc_control.lerc_api').setLevel(logging.WARNING)

        if args.write_template:
           import shutil
           shutil.copyfile(os.path.join(BASE_DIR, 'etc', 'example_remediate_routine.ini'), 'remediate.ini')
           print("Wrote remediate.ini")
           sys.exit(0)
        from lerc_control import remediate
        if args.remediation_file:
            remediate.Remediate(client, args.remediation_file)
        if args.kill_process_name:
            cmd = remediate.kill_process_name(client, args.kill_process_name)
            remediate.evaluate_remediation_results(cmd, 'process_names', args.kill_process_name)
        if args.kill_process_id:
            cmd = remediate.kill_process_id(client, args.kill_process_id)
            remediate.evaluate_remediation_results(cmd, 'pids', args.kill_process_id)
        if args.delete_registry_value:
            cmd = remediate.delete_registry_value(client, args.delete_registry_value)
            remediate.evaluate_remediation_results(cmd, 'registry_values', args.delete_registry_value)
        if args.delete_registry_key:
            cmd = remediate.delete_registry_key(client, args.delete_registry_key)
            remediate.evaluate_remediation_results(cmd, 'registry_keys', args.delete_registry_key)
        if args.delete_file:
            cmd = remediate.delete_file(client, args.delete_file)
            remediate.evaluate_remediation_results(cmd, 'files', args.delete_file)
        if args.delete_directory:
            cmd = remediate.delete_directory(client, args.delete_directory)
            remediate.evaluate_remediation_results(cmd, 'directories', args.delete_directory)
        if args.delete_scheduled_task:
            cmd = remediate.delete_scheduled_task(client, args.delete_scheduled_task)
            remediate.evaluate_remediation_results(cmd, 'scheduled_tasks', args.delete_scheduled_task)
        if args.delete_service:
            # delete_service returns a list of commands it issued
            cmds = remediate.delete_service(client, args.delete_service, auto_fill=True)
            for cmd in cmds:
                if isinstance(cmd, tuple):
                    remediate.evaluate_remediation_results(cmd[0], cmd[1], cmd[2])
                else:
                    remediate.evaluate_remediation_results(cmd, 'services', args.delete_service)
        sys.exit(0)

    # collections
    if args.instruction == 'collect':
        default_full_collect = True
        # From argparse Namespace, get collection script names if the argument is set
        collect_scripts = [carg for carg, value in vars(args).items() if carg.startswith('collect_') and value is True]

        if args.sub_collect_scipts:
            # If no specific scripts specified from this collect category, then we run all scripts
            if not any(args.sub_collect_scipts in script for script in collect_scripts):
                collect_scripts.extend([sname for sname in config['scripts'] if sname.startswith('collect_'+args.sub_collect_scipts)])

        if not args.debug:
            logging.getLogger('lerc_control.lerc_api').setLevel(logging.WARNING)
        if args.directory:
            commands = collect.get_directory(client, args.directory)
            default_full_collect = False
        if args.multi_collect:
            collect.multi_collect(client, args.multi_collect)
            default_full_collect = False
        if args.process_memory:
            collect.get_process_memory(client, args.process_memory)
            default_full_collect = False
        if args.proc_mem_strings:
            collect.get_process_memstrings(client, args.proc_mem_strings)
            default_full_collect = False
        if collect_scripts:
            cmds = []
            for script in collect_scripts:
                try:
                    cmds.extend(execute_script(client, config['scripts'][script], return_result_commands=True, execute_cleanup_commands=False))
                except:
                    sys.exit(1)
            cleanup_cmds = COMMON_CLEANUP_COMMANDS
            for cmd in cleanup_cmds['RUN']:
                cmd = client.Run(cmd)
                logger.info("Issued cleanup command {} : {}".format(cmd.id, cmd.command))
            written_cmd_results = get_script_results(cmds)
            for cmd in written_cmd_results:
                print("\t+ Results from CMD {} written: {}".format(cmd.id, cmd.write_results_path))
            if len(cleanup_cmds['QUIT']) > 0:
                cmd = client.Quit()
                logger.info("Issued command {} for client to uninstall itself from host.".format(cmd.id))
            default_full_collect = False
        # Any arguments where we turn on lerc_api info logger should go below
        if args.file_path:
            cmd = client.Upload(args.file_path)
            logger.info("Issued {} to collect file : {}".format(cmd.id, args.file_path))
            logging.getLogger('lerc_control.lerc_api').setLevel(logging.INFO)
            cmd.wait_for_completion()
            cmd.get_results()
            default_full_collect = False
        if args.reg_key_value:
            reg_value = args.reg_key_value[args.reg_key_value.rfind('\\')+1:].replace(' ','_')
            logging.getLogger('lerc_control.lerc_api').setLevel(logging.INFO)
            cmd = collect.collect_registry_key_value(client, args.reg_key_value)
            cmd.get_results(file_path="{}_{}_reg_{}.txt".format(client.hostname, cmd.id, reg_value))
            default_full_collect = False
        if args.reg_key:
            reg_key = args.reg_key[args.reg_key.rfind('\\')+1:].replace(' ','_')
            logging.getLogger('lerc_control.lerc_api').setLevel(logging.INFO)
            cmd = collect.collect_registry_key(client, args.reg_key)
            cmd.wait_for_completion()
            cmd.get_results(file_path="{}_{}_reg_{}.txt".format(client.hostname, cmd.id, reg_key))
            default_full_collect = False
        if default_full_collect:
            collect.full_collection(client)
        sys.exit(0)

    if args.instruction == 'script':
        config = ls.get_config
        if args.list_scripts:
            if not config.has_section('scripts'):
                print("\nNo pre-existing scripts have been made availble.")
                sys.exit(0)           
            print("\nAvailable scripts:")
            for sname in config['scripts']:
                print("\t{}".format(sname))
            print()
            sys.exit(0)
        elif args.script_name:
            if not config.has_option('scripts', args.script_name):
                logger.error("{} is not a defined script".format(args.script_name))
                sys.exit(1)
            script_path = config['scripts'][args.script_name]
            commands = execute_script(client, script_path)
            sys.exit(0)
        elif args.file_path:
            if not os.path.exists(args.file_path):
                logger.error("Could not find script file at '{}'".format(args.file_path))
                sys.exit(1)
            commands = execute_script(client, args.file_path)
            sys.exit(0)
        else:
            logger.info("No argument was specified for the script command. Exiting.")
            sys.exit(0)

    # Else, see if we're running a command directly
    cmd = None
    if args.instruction == 'run':
        if args.async:
            cmd = client.Run(args.command, async=args.async)
        else:
            cmd = client.Run(args.command)

    elif args.instruction == 'contain':
        if args.on:
            client.contain()
        elif args.off:
            client.release_containment()
        elif args.status:
            if not args.debug:
                logging.getLogger('lerc_control.lerc_api').setLevel(logging.WARNING)
            logger.info("Checking containment status for {}. This may take a moment ... ".format(client.hostname))
            containment_check = client.Run('ping google.com')
            firewall_status = client.Run('netsh advfirewall show allprofiles')
            containment_check.wait_for_completion()
            results = containment_check.get_results(return_content=True).decode('utf-8')
            if results and 'General failure.' in results:
                logger.info("{} is currently contained.".format(client.hostname))
            else:
                firewall_status.wait_for_completion()
                results = firewall_status.get_results(return_content=True).decode('utf-8')
                if results and 'AllowOutbound' in results:
                    logger.info("{} is NOT contained.".format(client.hostname))
                else:
                    logger.info("Unable to determine containment status. Printing firewall status:")
                    logger.info(results)

    elif args.instruction == 'download':
        # if client_file_path is not specified the client will write the file to it's local dir
        analyst_file_path = os.path.abspath(args.file_path)
        file_name = args.file_path[args.file_path.rfind('/')+1:]
        if args.local_file is None:
            args.local_file = file_name
        cmd = client.Download(file_name, client_file_path=args.local_file, analyst_file_path=analyst_file_path)

    elif args.instruction == 'upload':
        cmd = client.Upload(args.file_path)

    elif args.instruction == 'quit':
        cmd = client.Quit()
    elif args.check:
        command = ls.get_command(args.check)
        print(command)
        sys.exit()
    elif args.get:
        command = ls.get_command(args.get)
        if command:
            command.get_results(chunk_size=16384)
            print(command)
        sys.exit()
    elif args.resume:
        command = ls.get_command(args.resume)
        command.wait_for_completion()
        if command:
            print(command)
        sys.exit()
    else:
        print(client)
        sys.exit()

    if not cmd:
        sys.exit(1)

    if not cmd.wait_for_completion():
        logger.warning("{} (ID:{}) command went to a {} state. Exiting.".format(cmd.operation, cmd.id, cmd.status))
        sys.exit(1)
    logger.info("{} command {} completed successfully".format(cmd.operation, cmd.id))
    content = None
    if args.instruction == 'run':
        if args.print_only:
            content = cmd.get_results(return_content=args.print_only)
            print(content.decode('utf-8'))
        elif args.write_only:
            cmd.get_results(print_run=False, file_path=args.output_filename)
        else:
            cmd.get_results(file_path=args.output_filename)
    else:
        cmd.get_results()

    print(cmd)

    sys.exit()





import os
import time
import subprocess
import shlex
import pprint
import logging
import configparser

from lerc_control.scripted import load_script, execute_script
from lerc_control import lerc_api

logger = logging.getLogger("lerc_control."+__name__)

# Get the working lerc_control directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

def get_process_memory(lerc, pid, script_path='scripts/get_process_memory.ini'):
    return execute_script(lerc, script_path, placeholders={'PID': pid})

def get_process_memstrings(lerc, pid, script_path='scripts/get_process_memstrings.ini'):
    return execute_script(lerc, script_path, placeholders={'PID': pid})

def get_directory(lerc, dir_path, async_run=False):
    # TODO XXX Convert this into a script
    """Compress a directory with 7zip and upload the compressed result.

    :param lerc_api.Client lerc: A lerc_api.Client
    :param str dir_path: The path to the client directory
    :return: list of lerc_api.Command objects on success, else False
    """
    if not isinstance(lerc, lerc_api.Client):
        logger.error("Argument is of type:{} instead of type lerc_api.Client".format(type(lerc)))
        return False

    config = lerc._ls.get_config
    required_keys=['7za_path', '7za_dir_cmd']
    if not lerc_api.check_config(config, required_keys=required_keys):
        logger.error("^ Missing required config items")
        return False
    profile = lerc._ls.profile
    section = profile + "_collect"
    _7za_path = config[section]['7za_path'] if config.has_option(section, '7za_path') else config['default_collect']['7za_path']
    _7za_cmd = config[section]['7za_dir_cmd'] if config.has_option(section, '7za_dir_cmd') else config['default_collect']['7za_dir_cmd']

    if not os.path.exists(_7za_path):
        if _7za_path[0] == '/':
            _7za_path = BASE_DIR + _7za_path
        else:
            _7za_path = BASE_DIR + '/' + _7za_path

    if not os.path.exists(_7za_path):
        logger.error("'{}' does not exist".format(_7za_path))
        return False

    commands = []
    cmd = lerc.Download(_7za_path)
    logger.info("Issued CID={} to download 7za.exe.".format(cmd.id))
    commands.append(cmd)

    cmd = lerc.Run(_7za_cmd.format(lerc.hostname+'_dirOfInterest' , dir_path))
    logger.info("Issued CID={} to run 7za on '{}'".format(cmd.id, dir_path))
    commands.append(cmd)

    outputfile = '{}_dirOfInterest.7z'.format(lerc.hostname)
    upCmd = lerc.Upload(outputfile)
    logger.info("Issued CID={} to upload {}_dirOfInterest.7z".format(upCmd.id, lerc.hostname))
    logger.info("Waiting for the upload command to reach completion ... ")
    commands.append(upCmd.wait_for_completion())

    cmd = lerc.Run('Del "{}" && Del 7za.exe'.format(outputfile))
    logger.info("Issued CID={} to to delete '{}' and 7za.exe".format(cmd.id, outputfile))
    commands.append(cmd)

    if async_run:
        return upCmd

    logger.info("Getting result from the control server..".format(lerc.hostname))
    if upCmd.get_results(file_path='{}_dirOfInterest.7z'.format(lerc.hostname)):
        logger.info("Wrote {}_dirOfInterest.7z".format(lerc.hostname))
    return commands

def collect_registry_key(client, reg_path):
    """Collect all of the values at this registry key.

    :param str reg_path: Path the reg key
    :param lerc_api.Client client: A lerc client object to work with.
    """
    cmd = client.Run('reg query "{}"'.format(reg_path))
    logger.info("Issued command:{} to '{}' on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    return cmd

def collect_registry_key_value(client, reg_path):
    """Collet the registry data at the specific registry path key value. 

    :param str reg_path: Registry key path with specific key value appended like so: path-to-key+"\\key_value"
    :param lerc_api.Client client: A lerc client object to work with.
    """
    reg_key = reg_path[reg_path.rfind('\\')+1:]
    reg_path = reg_path[:reg_path.rfind('\\')]
    cmd = 'reg query "{}" /v "{}"'.format(reg_path, reg_key)
    cmd = client.Run(cmd)
    logger.info("Issued command:{} to '{}' on LERC:{}".format(cmd.id, cmd.command, client.hostname))
    return cmd 


def multi_collect(client, collect_file):
    """Pass a config formatted file describing artifacts to collect from the client.

    """
    if not os.path.exists(collect_file):
        logger.error("'{}' does not exist.".format(collect_file))
        return False

    config = configparser.ConfigParser()
    config.read(collect_file) 

    files = config['files']
    dirs = config['directories']
    regValues = config['registry_values']
    regKeys = config['registry_keys']

    commands = {'files': [],
                'directories': [],
                'registry_values': [],
                'registry_keys': []}

    for f in files:
        cmd = client.Upload(files[f])
        logger.info("Issued command:{} to upload '{}' from LERC:{}".format(cmd.id, files[f], client.hostname))
        commands['files'].append(cmd)

    for reg in regValues: 
        commands['registry_values'].append(collect_registry_key_value(client, regValues[reg]))

    for reg in regKeys:
        commands['registry_keys'].append(collect_registry_key(client, regKeys[reg]))

    for d in dirs:
        commands['directories'].append(get_directory(client, dirs[d], async_run=True))

    for cmd in commands['files']:
        logger.info("Waiting for command {} to complete..".format(cmd.id)) 
        cmd.wait_for_completion()
        logger.info("Getting results for '{}' collection.".format(cmd.client_file_path))
        filename = cmd.client_file_path[cmd.client_file_path.rfind('\\')+1:]
        filename = filename.replace(' ', '_')
        if cmd.get_results(file_path='{}_{}_file_{}'.format(client.hostname, cmd.id, filename)):
            logger.info("+ wrote '{}_{}_file_{}'".format(client.hostname, cmd.id, filename))

    for cmd in commands['registry_values']:
        rvalue = [r for r in [regValues[r] for r in regValues] if r[r.rfind('\\')+1:] in cmd.command and r[:r.rfind('\\')] in cmd.command][0]
        logger.info("Waiting for command {} to complete..".format(cmd.id))
        cmd.wait_for_completion()
        logger.info("Getting results of '{}' collections..".format(rvalue))
        rvalue_name = rvalue[rvalue.rfind('\\')+1:]
        rvalue_name = rvalue_name.replace(' ','_')
        if cmd.get_results(print_run=False, file_path='{}_{}_reg_{}'.format(client.hostname, cmd.id, rvalue_name)):
            logger.info("+ wrote '{}_{}_reg_{}'".format(client.hostname, cmd.id, rvalue_name))

    for cmd in commands['registry_keys']:
        rkey = [r for r in [regKeys[r] for r in regKeys] if r[r.rfind('\\')+1:] in cmd.command and r[:r.rfind('\\')] in cmd.command][0]
        logger.info("Waiting for command {} to complete..".format(cmd.id))
        cmd.wait_for_completion()
        logger.info("Getting results of '{}' collections..".format(rkey))
        rkey_name = rkey[rkey.rfind('\\')+1:]
        if cmd.get_results(print_run=False, file_path='{}_{}_reg_{}'.format(client.hostname, cmd.id, rkey_name)):
            logger.info("+ wrote '{}_{}_reg_{}'".format(client.hostname, cmd.id, rkey_name))

    for cmd in commands['directories']:
        directory = [d for d in [dirs[d] for d in dirs] if d in cmd.command][0]
        logger.info("Waiting for command {} to complete..".format(cmd.id))
        cmd.wait_for_completion()
        dirname = directory[directory.rfind('\\')+1:].replace(' ','_')
        logger.info("Getting results of '{}' collection..".format(cmd.client_file_path))
        if cmd.get_results(file_path='{}_{}_dir_{}.7z'.format(client.hostname, cmd.id, dirname)):
            logger.info("+ wrote '{}_{}_dir_{}'".format(client.hostname, cmd.id, dirname))

    return True

def full_collection(lerc):
    #########################################################################################
    ### This is a custom module built for a private collection package.                   ###
    ### :param lerc_api.Client lerc: A lerc Client object                                 ###
    #########################################################################################
    if not isinstance(lerc, lerc_api.Client):
        logger.error("Argument is of type:{} instead of type lerc_api.Client".format(type(lerc)))
        return False

    # Config config items exist and get them
    config = lerc._ls.get_config
    required_keys = ['lr_path', 'extract_cmd', 'collect_cmd', 'output_dir', 'streamline_path', 'client_working_dir']
    if not lerc_api.check_config(config, required_keys=required_keys):
        logger.error("^ Missing required configuration item(s)")
        return False
    profile = lerc._ls.profile
    collect_profile = profile+"_collect"
    client_workdir = config[profile]['client_working_dir']
    lr_path = config[collect_profile]['lr_path']
    extract_cmd = config[collect_profile]['extract_cmd']
    collect_cmd = config[collect_profile]['collect_cmd']
    output_dir = config[collect_profile]['output_dir']
    streamline_path = config[collect_profile]['streamline_path']

    commands = []

    logger.info("Starting full Live Response collection on {}.".format(lerc.hostname))

    # for contriving the output filename
    local_date_str_cmd = lerc.Run('date /t')
    # Delete any existing LR artifacts
    lerc.Run("DEL /S /F /Q lr && rmdir /S /Q lr")
    # download the package
    lr_download = lerc.Download(lr_path)
    logger.info("Issued CID={} for client to download {}.".format(lr_download.id, lr_path))
    # extract the package
    result = lerc.Run(extract_cmd)
    logger.info("Issued CID={} to extract lr.exe on the host.".format(result.id))
    # run the collection
    collect_command = lerc.Run(collect_cmd)
    logger.info("Issued CID={} to run {}.".format(collect_command.id, collect_cmd))
    # finish contriving the output filename
    output_filename = upload_command = None
    local_date_str_cmd.refresh()
    while True:
        if local_date_str_cmd.status == 'COMPLETE':
            dateStr = local_date_str_cmd.get_results(return_content=True).decode('utf-8')
            logger.debug("Got date string of '{}'".format(dateStr))
            try:
                # Mon 11/19/2018 -> 20181119 
                dateStr = dateStr.split(' ')[1].split('/')
                dateStr =  dateStr[2]+dateStr[0]+dateStr[1]
                 # hostname.upper() because streamline.py expects uppercase
                output_filename = lerc.hostname.upper() + "." + dateStr + ".7z"
            except IndexError:
                logger.warning("Unexpected date string obtained. Got '{}'".format(dateStr))
            break
        # wait five seconds before asking the server again
        time.sleep(5)
        local_date_str_cmd.refresh()
    # if output_filename is still None then an IndexError occured above, go a little more brute.
    if output_filename is not None:
        upload_command = lerc.Upload(client_workdir + output_dir + output_filename)
        logger.info("Issued CID={} to upload output at: '{}'".format(upload_command.id, client_workdir + output_dir + output_filename))
    # Stream back collect.bat output as it comes in
    logger.info("Streaming collect.bat output ... ")
    position = 0
    while True:
        collect_command.refresh()
        if collect_command.status == 'STARTED':
            if collect_command.filesize > 0:
                results = collect_command.get_results(return_content=True, position=position)
                if len(results) > 0:
                    position += len(results)
                    print(results.decode('utf-8'))
            time.sleep(1)
        elif collect_command.status == 'COMPLETE':
            if position < collect_command.filesize:
                results = collect_command.get_results(return_content=True, position=position)
                if len(results) > 0:
                    position += len(results)
                    print(results.decode('utf-8'))
            break
        elif collect_command.status == 'UNKNOWN' or collect_command.status == 'ERROR':
            logger.error("Collect command went to {} state : {}".format(collect_command.status, collect_command))
            return False
        time.sleep(5)
    if upload_command is None:
        output_dir_dict = lerc.list_directory(client_workdir + output_dir)
        dir_dict = output_dir_dict['dir_dict']
        # there should only be one file left in the dir
        dfiles = [item['name'] for item in dir_dict if item['type'] == 'file']
        if len(dfiles) > 1:
            logger.warning("More than one file found in output directory: {}".format(dfiles))
            logger.info("Issuing commands to collect all files.. ")
            for df in dfiles:
                upload_command = lerc.Upload(client_workdir + output_dir + df)
                logger.info("Issued CID={} to upload output at: '{}'".format(upload_command.id, client_workdir + output_dir + df))
                output_filename = df
        else:
            upload_command = lerc.Upload(client_workdir + output_dir + dfiles[0])
            logger.info("Issued CID={} to upload output at: '{}'".format(upload_command.id, client_workdir + output_dir + dfiles[0]))
            output_filename = dfiles[0]
    logger.info("Waiting for '{}' upload to complete.".format(output_filename))
    upload_command.wait_for_completion()
    
    if upload_command.status == 'COMPLETE':
        logger.info("Upload command complete. Telling lerc to delete the output file on the client")
        commands.append(lerc.Run('DEL /S /F /Q "{}"'.format(client_workdir + output_dir + output_filename)))

    # finally, stream the collection from the server to the cwd
    logger.info("Streaming {} from server..".format(output_filename))
    upload_command.get_results(file_path=output_filename)
    # Call steamline on the 7z lr package
    logger.info("[+] Starting streamline on {}".format(output_filename))
    args = shlex.split(streamline_path + " " + output_filename)
    try:
        subprocess.Popen(args).wait()
        logger.info("[+] Streamline complete")
    except Exception as e:
        logger.error("Exception with Streamline: {}".format(e))

    return True




import os
import logging
import pprint

from configparser import ConfigParser
from lerc_control import lerc_api

# GLOBALS #
logger = logging.getLogger("lerc_control."+__name__)

IGNORED_SECTIONS = ['overview']
REQUIRED_CMD_KEYS = ['operation']
OPTIONAL_CMD_KEYS = ['wait_for_completion', 'get_results']

REQUIRED_OP_KEY_MAP = {'RUN': ['command'],
                       'UPLOAD': ['path'],
                       'DOWNLOAD': ['file_path'],
                       'QUIT': []}
OPTIONAL_OP_KEY_MAP = {'RUN': ['async_run', 'write_results_path', 'print_results',
                               'common_setup_command', 'common_cleanup_command'],
                       'UPLOAD': ['write_results_path'],
                       'DOWNLOAD': ['client_file_path', 'common_setup_command'],
                       'QUIT': ['common_cleanup_command']}

# Get the working lerc_control directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

COMMON_SETUP_COMMANDS = {'RUN': [],
                         'DOWNLOAD': []}
COMMON_CLEANUP_COMMANDS = {'RUN': [],
                           'QUIT': []}

def script_missing_required_keys(config, KEYS):
    for key in KEYS:
        for section in config.sections():
            if section in IGNORED_SECTIONS:
                return False
            if not config.has_option(section, key):
                logger.error("{} is missing required key: {}".format(section, key))
                return True
    return False

def operation_missing_required_keys(config, section, KEYS):    
    for key in KEYS:
        if not config.has_option(section, key):
            logger.error("{} is missing required operation key:{} for operation:{}".format(section, key, config[section]['operation']))
            return True
    return False


def get_script_results(cmds):
    """
    Wait for and collect results from the given list of script commands.
    """
    written_result_cmds = []
    for cmd in cmds:
        logger.info("Waiting for command {} to complete..".format(cmd.id))
        cmd.wait_for_completion()
        logger.info("Getting the results for command {}".format(cmd.id))
        cmd.get_results(file_path=cmd.write_results_path, print_run=cmd.print_results)
        if cmd.write_results_path and os.path.exists(cmd.write_results_path):
            logger.info("Wrote results: {}".format(cmd.write_results_path))
            written_result_cmds.append(cmd)
    return written_result_cmds


def load_script(script_path):
    script = ConfigParser()
    if not os.path.exists(script_path):
        if script_path[0] == '/':
            script_path = BASE_DIR + script_path
        else:
            script_path = BASE_DIR + '/' + script_path
    if not os.path.exists(script_path):
        script_name = script_path[script_path.rfind('/')+1:script_path.rfind('.')]
        logger.error("The path to the script named '{}' does not exist.".format(script_name))
        return False
    try:
        script.read(script_path)
    except Exception as e:
        logger.error("ConfigParser Error reading '{}' : {}".format(script_path, e))
        return False
    if script_missing_required_keys(script, REQUIRED_CMD_KEYS):
        return False
    return script

def enforce_argument_placeholders(script, placeholders={}):
    """Some scripts require arguments to execute. These arguments are in the form of string format placeholder keys.
    Make sure we have all of the arguments for this script. Prompt the user for those arguments if we don't.

    :param ConfigParser script: A loaded lerc script
    :param dict placeholders: A dictionary of placeholders we already have (if any)
    :return: placeholders
    """
    logger.debug("Enforcing argument placeholders")
    args_needed = []
    required_args = script['overview']['required_arguments'].split(',') if script.has_option('overview', 'required_arguments') else None
    if required_args:
        args_needed = [arg for arg in required_args if arg not in placeholders.keys()]
    for arg in args_needed:
        placeholders[arg] = input("Script needs argument {}:".format(arg))
    return placeholders
    # leaving this here in case I decide to do the enforcement based on {whatever} is in certain command fields
    #fields = [i[1] for i in Formatter().parse(write_results_path) if i[1] is not None]


def execute_script(lerc, script_path, return_result_commands=False, execute_cleanup_commands=True, placeholders={}):
    """Execute a script on this host.

    :param lerc_api.Client lerc: A lerc_api.Client object.
    :param str script_path: the path to the script
    :param bool return_result_commands: If True, return list of commands that we need to get results from.
    :param bool execute_cleanup_commands: common_cleanup_commands will be executed if True. Else they're stored in COMMON_CLEANUP_COMMANDS
    :return: a dictionary of the commands issued
    """

    if not isinstance(lerc, lerc_api.Client):
        logger.error("Argument is of type:{} instead of type lerc_api.Client".format(type(lerc)))
        return False
 
    script = load_script(script_path)
    if 'HOSTNAME' not in placeholders:
        # This is a generic, common placeholder for writing result files
        placeholders['HOSTNAME'] = lerc.hostname
    placeholders = enforce_argument_placeholders(script, placeholders)

    command_history = {}
    commands = [cmd for cmd in script.sections() if cmd not in IGNORED_SECTIONS]

    script_name = script_path[script_path.rfind('/')+1:script_path.rfind('.')]
    # make sure requirements are met first
    for command in commands:
        op =  script[command]['operation'].upper()
        if op not in REQUIRED_OP_KEY_MAP:
            logger.error("{} is not a recognized lerc operation!".format(op))
            return False
        if operation_missing_required_keys(script, command, REQUIRED_OP_KEY_MAP[op]):
            return False

    logger.info("Beginning execution of {}".format(script_name))
    for command in commands:
        logger.info("Processing {}".format(command))
        command_history[command] = {}
        op =  script[command]['operation'].upper()

        print_results = True
        get_results = False
        write_results_path = None
        is_common_setup_command = False
        is_common_cleanup_command = False
        message = False
        if 'get_results' in script[command]:
            get_results = script[command].getboolean('get_results')
        if 'write_results_path' in script[command]:
            write_results_path = script[command]['write_results_path'].format(**placeholders)
        if 'common_setup_command' in script[command]:
            is_common_setup_command = script[command].getboolean('common_setup_command')
        if 'common_cleanup_command' in script[command]:
            is_common_cleanup_command = script[command].getboolean('common_cleanup_command')
        if 'message' in script[command]:
            message = script[command]['message'].format(**placeholders)

        if op == 'RUN':
            async_run = False
            if 'async_run' in script[command]:
                async_run = script[command].getboolean('async_run')
            if 'print_results' in script[command]:
                print_results = script[command].getboolean('print_results')
            run_string = script[command]['command'].format(**placeholders)
            if is_common_setup_command:
                cmd = [cmd for cmd in COMMON_SETUP_COMMANDS['RUN'] if run_string == cmd.command]
                if cmd:
                    if len(cmd) != 1:
                        logger.error("Logic error - more than one common command with the same setup command: {}".format(cmd))
                    cmd = cmd[0]
                    logger.info('Skipping common setup command for script:{} as it should already be satisfied by: CMD={} -> {}'.format(script_name, cmd.id, cmd.command))
                    continue
            if is_common_cleanup_command and not execute_cleanup_commands:
                cmd = [cmd_string for cmd_string in COMMON_CLEANUP_COMMANDS['RUN'] if run_string == cmd_string]
                if not cmd:
                    # assuming we never will want to async_run OR write_results_path OR print_results
                    COMMON_CLEANUP_COMMANDS['RUN'].append(run_string)
                continue
            cmd = lerc.Run(run_string, async=async_run)
            command_history[command] = cmd
            command_history[command].get_the_results = get_results
            command_history[command].write_results_path = write_results_path
            command_history[command].print_results = print_results
            command_history[command].is_common_setup_command = is_common_setup_command
            if is_common_setup_command:
                COMMON_SETUP_COMMANDS['RUN'].append(cmd)
            logger.info("Issued : Run - CID={} - {}".format(cmd.id, run_string))
            if message:
                logger.info("SCRIPT MESSAGE: {}".format(message))
        elif op == 'DOWNLOAD':
            client_file_path = None
            if 'client_file_path' in script[command]:
                client_file_path = script[command]['client_file_path'].format(**placeholders)
            file_path = script[command]['file_path'].format(**placeholders)
            if not os.path.exists(file_path):
                old_fp = file_path
                if file_path[0] == '/':
                    file_path = BASE_DIR + file_path
                else:
                    file_path = BASE_DIR + '/' + file_path
                if not os.path.exists(file_path):
                    logger.error("Not found: '{}' OR '{}'".format(old_fp, file_path))
                    return False
            if is_common_setup_command:
                cmd = [cmd for cmd in COMMON_SETUP_COMMANDS['DOWNLOAD'] if file_path == cmd.analyst_file_path]
                if client_file_path and cmd:
                    cmd2 = [cmd for cmd in COMMON_SETUP_COMMANDS['DOWNLOAD'] if client_file_path == cmd.client_file_path]
                    if cmd2 and cmd[0].id != cmd2[0].id:
                        logger.debug("Similar but different setup commands issued previously.")
                        # we should issue a new command
                        cmd = None
                if cmd:
                    if len(cmd) != 1:
                        logger.warning("Logic error - more than one common command with the same setup command: {}".format(cmd))
                    cmd = cmd[0]
                    logger.info('Skipping common setup command for script:{} as it should already be satisfied by: CMD={} to download {}'.format(script_name, cmd.id, cmd.analyst_file_path))
                    continue
            cmd = lerc.Download(file_path, client_file_path=client_file_path)
            command_history[command] = cmd
            command_history[command].is_common_setup_command = is_common_setup_command
            if is_common_setup_command:
                COMMON_SETUP_COMMANDS['DOWNLOAD'].append(cmd)
            logger.info("Issued : Download - CID={} - {}".format(cmd.id, file_path))
            if message:
                logger.info("SCRIPT MESSAGE: {}".format(message))
        elif op == 'UPLOAD':
            path = script[command]['path'].format(**placeholders)
            cmd = lerc.Upload(path)
            command_history[command] = cmd
            command_history[command].get_the_results = get_results
            command_history[command].write_results_path = write_results_path
            command_history[command].print_results = False
            logger.info("Issued : Upload - CID={} - {}".format(cmd.id, path))
            if message:
                logger.info("SCRIPT MESSAGE: {}".format(message))
        elif op == 'QUIT':
            if is_common_cleanup_command and not execute_cleanup_commands:
                if not COMMON_CLEANUP_COMMANDS['QUIT']:
                    COMMON_CLEANUP_COMMANDS['QUIT'].append(True)
                continue
            cmd = lerc.Quit()
            command_history[command] = cmd
            logger.info("Issued : Quit - CID={}".format(cmd.id))
            if message:
                logger.info("SCRIPT MESSAGE: {}".format(message))

    logger.info("Checking to see if results need to be obtained ...")
    result_commands = []
    for command in command_history:
        cmd = command_history[command]
        if hasattr(cmd, 'get_the_results') and cmd.get_the_results:
            if return_result_commands:
                result_commands.append(cmd)
            else:
                logger.info("Waiting for command {} to complete..".format(cmd.id))
                cmd.wait_for_completion()
                logger.info("Getting the results for command {}".format(cmd.id))
                cmd.get_results(file_path=cmd.write_results_path, print_run=cmd.print_results)
                if cmd.write_results_path and os.path.exists(cmd.write_results_path):
                    logger.info("Wrote results: {}".format(cmd.write_results_path))

    if return_result_commands:
        return result_commands
    return command_history

# XXX Maybe create a LERC Script class??